/*
=========================================================================
rtl8169.c: A RealTek Gigabit Ethernet driver for Linux kernel 2.2.x.
 --------------------------------------------------------------------

 History:
 Feb 4 2002, ShuChen - created initially.

 --------------------------------------------------------------------

 Copyright 2002 by Realtek Semiconductor Corp.
=========================================================================
 The link speed and duplex mode can be forced in 5 modes.
	 Command:	"insmod rtl8169 media = SET_MEDIA"
	 Ex:		"insmod rtl8169 media = 0x04" will force link mode to be 100Mpbs Half-duplex.

	The value of SET_MEDIA can be:
 		_10_Half	= 0x01
 		_10_Full	= 0x02
 		_100_Half	= 0x04
 		_100_Full	= 0x08
 		_1000_Full	= 0x10

=========================================================================
RTL8169_VERSION "1.1"	<2002/10/4>

	The bit4:0 of MII register 4 is called "selector field", and have to be
	00001b to indicate support of IEEE std 802.3 during NWay process of
	exchanging Link Code Word (FLP).

RTL8169_VERSION "1.2"	<2003/09/01>
	Update driver module name.
	Modify ISR.
	Add chip mac_version.
	Add chip phy_version.
	Add priv->phy_timer_t, rtl8169_phy_timer_t_handler()
	Add rtl8169_hw_PHY_config()
	Add rtl8169_hw_PHY_reset()
	Add tx_bytes, rx_bytes.
	Set 0x0000 to PHY at offset 0x0b.
	Modify chip mac_version, phy_version
	Force media for multiple card.
	Modify receive data buffer.
*/


#define RTL8169_VERSION "v1.2 <2003-09-01>"
#define MODNAME "RTL8169s/8110s"
#define RTL8169_DRIVER_NAME   MODNAME " Gigabit Ethernet driver on kernel 2.2.x " RTL8169_VERSION


#undef RTL8169_DEBUG
//#define RTL8169_DEBUG


#ifdef RTL8169_DEBUG
#  define assert(expr) \
        if(!(expr)) {                                   \
                printk( "Assertion failed! %s,%s,%s,line=%d\n", #expr,__FILE__,__FUNCTION__,__LINE__);          \
        }
#  define DBG_PRINT( fmt, args...)   printk(KERN_INFO "r8169: " fmt, ## args);
#else
#  define assert(expr) do {} while (0)
#  define DBG_PRINT( fmt, args...)   do{}while(0);
#endif  // end of #ifdef RTL8169_DEBUG



#ifndef USE_MEM_OPS
#define USE_IO_OPS
#endif


/* media options */
#define MAX_UNITS 8
static int media[MAX_UNITS] = {-1, -1, -1, -1, -1, -1, -1, -1};


/* Maximum events (Rx packets, etc.) to handle at each interrupt. */
static int max_interrupt_work = 20;

/* Maximum number of multicast addresses to filter (vs. Rx-all-multicast).*/
static int multicast_filter_limit = 32;

/* debug flag */
static int debug = 0;


/* Operational parameters that usually are not changed. */
/* Time in jiffies before concluding the transmitter is hung. */
#define TX_TIMEOUT  (6*HZ)

#ifndef __KERNEL__
#define __KERNEL__
#endif


#include <linux/config.h>
#include <linux/version.h>
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/string.h>
#include <linux/timer.h>
#include <linux/errno.h>
#include <linux/ioport.h>
#include <linux/malloc.h>
#include <linux/interrupt.h>
#include <linux/pci.h>
#include <linux/netdevice.h>
#include <linux/etherdevice.h>
#include <linux/skbuff.h>
#include <asm/processor.h>		/* Processor type for cache alignment. */
#include <asm/io.h>

#if LINUX_VERSION_CODE >= 0x20300
#include <linux/spinlock.h>
#elif LINUX_VERSION_CODE >= 0x20200
#include <asm/spinlock.h>
#endif


#include "kern_compat.h"



static void *rtl8169_probe1(struct pci_dev *pdev, void *init_dev, long ioaddr, int irq, int chip_idx, int find_cnt);

enum chip_capability_flags {HAS_MII_XCVR=0x01, HAS_CHIP_XCVR=0x02, HAS_LNK_CHNG=0x04};

#ifdef USE_IO_OPS
#  define RTL8169_IOTYPE  PCI_USES_MASTER|PCI_USES_IO |PCI_ADDR0
#else
#  define RTL8169_IOTYPE  PCI_USES_MASTER|PCI_USES_MEM|PCI_ADDR1
#endif

#define RTL8169_CAPS  HAS_CHIP_XCVR|HAS_LNK_CHNG


static struct pci_id_info pci_tbl[] = {
//	{"RealTek RTL8169 Fast Ethernet on kernel 2.2.x", { 0x816910ec, 0xffffffff,}, RTL8169_IOTYPE, 0x80, RTL8169_CAPS, },
	{ RTL8169_DRIVER_NAME, { 0x816910ec, 0xffffffff,}, RTL8169_IOTYPE, 0x80, RTL8169_CAPS, },
	{0,},						/* 0 terminated list. */
};


struct drv_id_info rtl8169_drv_id = {
	"realtek", PCI_HOTSWAP, PCI_CLASS_NETWORK_ETHERNET<<8, pci_tbl,
	rtl8169_probe1, /*rtl_pwr_event*/ NULL };


#ifndef USE_IO_OPS
#undef inb
#undef inw
#undef inl
#undef outb
#undef outw
#undef outl
#define inb readb
#define inw readw
#define inl readl
#define outb writeb
#define outw writew
#define outl writel
#endif


#define RTL_R8(reg)         inb(ioaddr + reg)
#define RTL_R16(reg)        inw(ioaddr + reg)
#define RTL_R32(reg)        inl(ioaddr + reg)
#define RTL_W8(reg,val8)    outb((val8), ioaddr + reg)
#define RTL_W16(reg,val16)  outw((val16), ioaddr + reg)
#define RTL_W32(reg,val32)  outl((val32), ioaddr + reg)


#define RTL_GIGA_MAC_VER_B  0x00    //MAC version 0000
//#define RTL_GIGA_MAC_VER_C  0x03
#define RTL_GIGA_MAC_VER_D  0x01    //MAC version 0001
#define RTL_GIGA_MAC_VER_E  0x02    //MAC version 0002

#define RTL_GET_MAC_VERSION(mac_version)     \
{\
	mac_version = RTL_GIGA_MAC_VER_B;\
	if( (RTL_R32(TxConfig)&0x7c800000) & (0x1<<26) ){ \
		mac_version = RTL_GIGA_MAC_VER_E;\
	}\
	else if( (RTL_R32(TxConfig)&0x7c800000) & (0x1<<23) ){\
		mac_version = RTL_GIGA_MAC_VER_D;\
	}\
	else if( (RTL_R32(TxConfig)&0x7c800000) == 0x00000000 ){\
		mac_version = RTL_GIGA_MAC_VER_B;\
	}\
}

#define RTL_PRINT_MAC_VERSION(mac_version) \
{\
	switch(mac_version) \
	{ \
		case RTL_GIGA_MAC_VER_E: \
			DBG_PRINT("mac_version == RTL_GIGA_MAC_VER_E (0002)\n"); \
			break; \
		case RTL_GIGA_MAC_VER_D: \
			DBG_PRINT("mac_version == RTL_GIGA_MAC_VER_D (0001)\n"); \
			break; \
		case RTL_GIGA_MAC_VER_B: \
			DBG_PRINT("mac_version == RTL_GIGA_MAC_VER_B (0000)\n"); \
			break; \
		default: \
			DBG_PRINT("mac_version == Unknown\n"); \
			break; \
	} \
}

#define RTL_GIGA_PHY_VER_C    0x03	//PHY Reg 0x03 bit0-3 == 0x0000
#define RTL_GIGA_PHY_VER_D    0x04	//PHY Reg 0x03 bit0-3 == 0x0000
#define RTL_GIGA_PHY_VER_E    0x05	//PHY Reg 0x03 bit0-3 == 0x0000
#define RTL_GIGA_PHY_VER_F    0x06	//PHY Reg 0x03 bit0-3 == 0x0001
#define RTL_GIGA_PHY_VER_G    0x07	//PHY Reg 0x03 bit0-3 == 0x0002

#define RTL_GET_PHY_VERSION(phy_version)     \
{\
	phy_version = RTL_GIGA_PHY_VER_D;\
	if( (RTL8169_READ_GMII_REG(ioaddr,3)&0x000f) == 0x0002 ){ \
		phy_version = RTL_GIGA_PHY_VER_G;\
	} \
	else if( (RTL8169_READ_GMII_REG(ioaddr,3)&0x000f) == 0x0001 ){ \
		phy_version = RTL_GIGA_PHY_VER_F;\
	} \
	else if( (RTL8169_READ_GMII_REG(ioaddr,3)&0x000f) == 0x0000 ){ \
		phy_version = RTL_GIGA_PHY_VER_E;\
	} \
}

#define RTL_PRINT_PHY_VERSION(phy_version)  \
{\
	switch(phy_version)\
	{ \
		case RTL_GIGA_PHY_VER_G: \
			DBG_PRINT("phy_version == RTL_GIGA_PHY_VER_G (0002)\n"); \
			break; \
		case RTL_GIGA_PHY_VER_F: \
			DBG_PRINT("phy_version == RTL_GIGA_PHY_VER_F (0001)\n"); \
			break; \
		case RTL_GIGA_PHY_VER_E: \
			DBG_PRINT("phy_version == RTL_GIGA_PHY_VER_E (0000)\n"); \
			break; \
		case RTL_GIGA_PHY_VER_D: \
			DBG_PRINT("phy_version == RTL_GIGA_PHY_VER_D (0000)\n"); \
			break; \
		default: \
			DBG_PRINT("phy_version == Unknown\n"); \
			break; \
	} \
}




#define MAC_ADDR_LEN	6		/* MAC address length*/

#define NUM_TX_DESC	64			/* Number of Tx descriptor registers for rtl8169*/
#define NUM_RX_DESC	64			/* Number of Rx descriptor registers for rtl8169*/
#define RX_BUF_SIZE	1536		/* Rx Buffer size */

#define TX_FIFO_THRESH 256			/* In bytes, rounded down to 32 byte units. */
#define RX_FIFO_THRESH	7			/* 7 means NO threshold, Rx buffer level before first PCI xfer.  */
#define RX_DMA_BURST	6			/* Maximum PCI burst, '6' is 1024 */
#define TX_DMA_BURST	6			/* Maximum PCI burst, '6' is 1024 */
#define EarlyTxThld 	0x3F		/* 0x3F means NO early transmit */
#define RxPacketMaxSize	0x0800		/* Maximum size supported is 16K-1 */
#define InterFrameGap	0x03		/* 3 means InterFrameGap = the shortest one */

#define MAX_RX_SKBDATA_SIZE 1600

typedef struct timer_list timer_t;


/* Symbolic offsets to registers. */
enum RTL8169_registers {
	MAC0 = 0x0,		/* Ethernet hardware address. */
	MAR0 = 0x8,		/* Multicast filter. */
	TxDescStartAddr	= 0x20,
	TxHDescStartAddr= 0x28,
	FLASH	= 0x30,
	ERSR	= 0x36,
	ChipCmd	= 0x37,
	TxPoll	= 0x38,
	IntrMask = 0x3C,
	IntrStatus = 0x3E,
	TxConfig = 0x40,
	RxConfig = 0x44,
	RxMissed = 0x4C,
	Cfg9346 = 0x50,
	Config0	= 0x51,
	Config1	= 0x52,
	Config2	= 0x53,
	Config3	= 0x54,
	Config4	= 0x55,
	Config5	= 0x56,
	MultiIntr = 0x5C,
	PHYAR	= 0x60,
	TBICSR	= 0x64,
	TBI_ANAR = 0x68,
	TBI_LPAR = 0x6A,
	PHYstatus = 0x6C,
	RxMaxSize = 0xDA,
	CPlusCmd = 0xE0,
	RxDescStartAddr	= 0xE4,
	EarlyTxThres	= 0xEC,
	FuncEvent	= 0xF0,
	FuncEventMask	= 0xF4,
	FuncPresetState	= 0xF8,
	FuncForceEvent	= 0xFC,		
};

enum InterruptStatusBits {
	SYSErr 		= 0x8000,
	PCSTimeout	= 0x4000,
	SWInt		= 0x0100,
	TxDescUnavail	= 0x80,
	RxFIFOOver 	= 0x40,
	RxUnderrun 	= 0x20,
	LinkChg 	= 0x20,
	RxOverflow 	= 0x10,
	TxErr 	= 0x08,
	TxOK 	= 0x04,
	RxErr 	= 0x02,
	RxOK 	= 0x01,
};

enum RxStatusDesc {
	RxRES = 0x00200000,
	RxCRC = 0x00080000,
	RxRUNT= 0x00100000,
	RxRWT = 0x00400000,
};


enum ChipCmdBits {
	CmdReset = 0x10,
	CmdRxEnb = 0x08,
	CmdTxEnb = 0x04,
	RxBufEmpty = 0x01,
};


enum Cfg9346Bits {
	Cfg9346_Lock = 0x00,
	Cfg9346_Unlock = 0xC0,
};


enum rx_mode_bits {
	AcceptErr = 0x20,
	AcceptRunt = 0x10,
	AcceptBroadcast = 0x08,
	AcceptMulticast = 0x04,
	AcceptMyPhys = 0x02,
	AcceptAllPhys = 0x01,
};


/* Bits in RxConfig. */
enum RxConfigBits {
	RxCfgFIFOShift = 13,
	RxCfgDMAShift = 8,
};

/* Bits in TxConfig. */
enum TxConfigBits {
	TxInterFrameGapShift = 24,
	TxDMAShift = 8,		/* DMA burst value (0-7) is shift this many bits */
};


enum rtl8169_PHYstatus {
	TBI_Enable	= 0x80,
	TxFlowCtrl	= 0x40,
	RxFlowCtrl	= 0x20,
	_1000bpsF	= 0x10,
	_100bps		= 0x08,
	_10bps		= 0x04,
	LinkStatus	= 0x02,
	FullDup		= 0x01,
};


enum GIGABIT_PHY_registers {
	PHY_CTRL_REG = 0,
	PHY_STAT_REG = 1,
	PHY_AUTO_NEGO_REG = 4,
	PHY_1000_CTRL_REG = 9,
};

enum GIGABIT_PHY_REG_BIT {
	//PHY_CTRL_REG = 0;
	PHY_Restart_Auto_Nego	= 0x0200,
	PHY_Enable_Auto_Nego	= 0x1000,

	//PHY_STAT_REG = 1;
	PHY_Auto_Neco_Comp	= 0x0020,

	//PHY_AUTO_NEGO_REG = 4;
	PHY_Cap_10_Half		= 0x0020,
	PHY_Cap_10_Full		= 0x0040,
	PHY_Cap_100_Half	= 0x0080,
	PHY_Cap_100_Full	= 0x0100,

	//PHY_1000_CTRL_REG = 9;
	PHY_Cap_1000_Full	= 0x0200,

	PHY_Cap_Null		= 0x0,
};


enum _MediaType{
	_10_Half	= 0x01,
	_10_Full	= 0x02,
	_100_Half	= 0x04,
	_100_Full	= 0x08,
	_1000_Full	= 0x10,
};

enum _TBICSRBit {
	TBILinkOK 	= 0x02000000,
};


enum _DescStatusBit {
	OWNbit	= 0x80000000,
	EORbit = 0x40000000,
	FSbit	= 0x20000000,
	LSbit	= 0x10000000,
};


struct TxDesc {
	u32		status;
	u32		vlan_tag;
	u32		buf_addr;
	u32		buf_Haddr;
};

struct RxDesc {
	u32		status;
	u32		vlan_tag;
	u32		buf_addr;
	u32		buf_Haddr;
};


#define PRIV_ALIGN	15 	/* Desired alignment mask */
struct rtl8169_private {

	//=============
	struct net_device *next_module;
	void *priv_addr;					/* Unaligned address for kfree */
	void *mmio_addr;							/* memory map physical address*/
	struct pci_dev *pci_dev;					/* Index of PCI device  */
	struct net_device_stats stats;				/* statistics of net device */
	unsigned int rx_config;
	spinlock_t lock;							/* spin lock flag */
	int chipset;
	int chip_id;
	int mac_version;
	int phy_version;
	timer_t phy_timer_t;
	unsigned long phy_link_down_cnt;
	unsigned long cur_rx;						/* Index into the Rx descriptor buffer of next Rx pkt. */
	unsigned long cur_tx;						/* Index into the Tx descriptor buffer of next Rx pkt. */
	unsigned long dirty_tx;
	unsigned long tx_full;				/* The Tx queue is full. */
	unsigned char	*TxDescArrays;				/* Index of Tx Descriptor buffer */
	unsigned char	*RxDescArrays;				/* Index of Rx Descriptor buffer */
	struct	TxDesc	*TxDescArray;				/* Index of 256-alignment Tx Descriptor buffer */
	struct	RxDesc	*RxDescArray;				/* Index of 256-alignment Rx Descriptor buffer */
	struct	sk_buff	*Tx_skbuff[NUM_TX_DESC];	/* Index of Transmit data buffer */
	struct	sk_buff	*Rx_skbuff[NUM_RX_DESC];	/* Receive data buffer */
	unsigned char   drvinit_fail;

};

#ifdef MODULE
#if LINUX_VERSION_CODE > 0x20115
MODULE_AUTHOR("Realtek");
MODULE_DESCRIPTION("RealTek RTL8169 Gigabit Ethernet driver");
MODULE_PARM (media, "1-" __MODULE_STRING(MAX_UNITS) "i");
#endif
#endif


static int rtl8169_open(struct net_device *dev);
static int rtl8169_init_ring (struct net_device *dev);
static void rtl8169_hw_start (struct net_device *dev);
static int rtl8169_start_xmit(struct sk_buff *skb, struct net_device *dev);
static void rtl8169_tx_interrupt (struct net_device *dev, struct rtl8169_private *priv, long ioaddr);
static void rtl8169_rx_interrupt (struct net_device *dev, struct rtl8169_private *priv, long ioaddr);
static void rtl8169_interrupt(int irq, void *dev_instance, struct pt_regs *regs);
static void rtl8169_tx_clear (struct rtl8169_private *tp);
static int rtl8169_close(struct net_device *dev);
static struct net_device_stats *rtl8169_get_stats(struct net_device *dev);
static inline u32 ether_crc(int length, unsigned char *data);
static void set_rx_mode(struct net_device *dev);
static void rtl8169_soft_reset(struct net_device *dev);

static void rtl8169_hw_PHY_config (struct net_device *dev);
static void rtl8169_hw_PHY_reset(struct net_device *dev);


static const unsigned int rtl8169_tx_config = (TX_DMA_BURST << TxDMAShift) | (InterFrameGap << TxInterFrameGapShift) ;
static const unsigned int rtl8169_rx_config = (RX_FIFO_THRESH << RxCfgFIFOShift) | (RX_DMA_BURST << RxCfgDMAShift) ;
static const u16 rtl8169_intr_mask = LinkChg | RxOverflow | RxFIFOOver | TxErr | TxOK | RxErr | RxOK ;


/* A list of all installed RTL8129 devices, for removing the driver module. */
static struct net_device *root_rtl8169_dev = NULL;


#define RTL8169_WRITE_GMII_REG_BIT( ioaddr, reg, bitnum, bitval )\
{ \
	int val; \
	if( bitval == 1 ){ val = ( RTL8169_READ_GMII_REG( ioaddr, reg ) | (bitval<<bitnum) ) & 0xffff ; } \
	else{ val = ( RTL8169_READ_GMII_REG( ioaddr, reg ) & (~(0x0001<<bitnum)) ) & 0xffff ; } \
	RTL8169_WRITE_GMII_REG( ioaddr, reg, val ); \
}




#ifdef RTL8169_DEBUG
unsigned long alloc_rxskb_cnt = 0;
#define RTL8169_ALLOC_RXSKB(bufsize) 	dev_alloc_skb(bufsize); alloc_rxskb_cnt ++ ;
#define RTL8169_FREE_RXSKB(skb) 	kfree_skb(skb); alloc_rxskb_cnt -- ;
#define RTL8169_NETIF_RX(skb) 		netif_rx(skb); alloc_rxskb_cnt -- ;
#else
#define RTL8169_ALLOC_RXSKB(bufsize) 	dev_alloc_skb(bufsize);
#define RTL8169_FREE_RXSKB(skb) 	kfree_skb(skb);
#define RTL8169_NETIF_RX(skb) 		netif_rx(skb);
#endif //end #ifdef RTL8169_DEBUG




//======================================================================================================
void udelay(int cnt)
{
	int i=cnt,j;
	while( i > 0 ){
		i--;
		for(j=1000;j>0;j--){}
	}
}





//======================================================================================================
static void rtl8169_soft_reset(struct net_device *dev)
{
	int	i;
	long	ioaddr = dev->base_addr;

	RTL_W8( ChipCmd, CmdReset );

	// Check that the chip has finished the reset.
	for (i = 10000; i > 0; i--){
		if ( RTL_R8(ChipCmd) == 0)break;
		else udelay(1);
	}
}






//=================================================================
//	PHYAR
//	bit		Symbol
//	31		Flag
//	30-21	reserved
//	20-16	5-bit GMII/MII register address
//	15-0	16-bit GMII/MII register data
//=================================================================
void RTL8169_WRITE_GMII_REG( long ioaddr, int RegAddr, int value )
{
	int	i;

	RTL_W32( PHYAR, (0x80000000 | (RegAddr&0xFF)<<16 | value) );

	udelay(20);

#if 0
	for( i = 2000; i > 0 ; i -- ){
		// Check if the RTL8169 has completed writing to the specified MII register
		if( ! ( RTL_R32(PHYAR) & 0x80000000 ) ){
			break;
		}
		else{
			udelay(100);
		}
	}// end of for() loop
#endif  //end #if 0

//	return ( i!=0 ) ? 1 : 0 ;
}

//=================================================================
int RTL8169_READ_GMII_REG( long ioaddr, int RegAddr )
{
	int i, value = -1;
	
	RTL_W32( PHYAR, (0x0 | (RegAddr&0xFF)<<16) );

	udelay(1000);

	for( i = 2000; i > 0 ; i -- ){
		// Check if the RTL8169 has completed retrieving data from the specified MII register
		if( RTL_R32(PHYAR) & 0x80000000 ){
			value = (int)( RTL_R32(PHYAR)&0xFFFF );
			break;
		}
		else{
			udelay(100);
		}
	}// end of for() loop
	return value;
}



//======================================================================================================
static void rtl8169_hw_PHY_reset(struct net_device *dev)
{
	int val, phy_reset_expiretime = 50;
	struct rtl8169_private *priv = dev->priv;
	long ioaddr = (long)(priv->mmio_addr);

	DBG_PRINT("%s: Reset RTL8169s PHY\n", dev->name);

	val = ( RTL8169_READ_GMII_REG( ioaddr, 0 ) | 0x8000 ) & 0xffff;
	RTL8169_WRITE_GMII_REG( ioaddr, 0, val );

	do //waiting for phy reset
	{
		if( RTL8169_READ_GMII_REG( ioaddr, 0 ) & 0x8000 ){
			phy_reset_expiretime --;
			udelay(100);
		}
		else{
			break;
		}
	}while( phy_reset_expiretime >= 0 );

}




//======================================================================================================
static void rtl8169_hw_PHY_config (struct net_device *dev)
{
	struct rtl8169_private *priv = dev->priv;
	long ioaddr = (long)(priv->mmio_addr);
	int val;

	RTL_PRINT_MAC_VERSION(priv->mac_version);
	RTL_PRINT_PHY_VERSION(priv->phy_version);

	if( (priv->mac_version > RTL_GIGA_MAC_VER_B) && ( priv->phy_version < RTL_GIGA_PHY_VER_F ) )
	{
			DBG_PRINT("MAC version!=0 && PHY version == 0 or 1\n");
			DBG_PRINT("Do final_reg2.cfg\n");
			// phy config for RTL8169s mac_version C chip
			RTL8169_WRITE_GMII_REG( ioaddr, 31, 0x0001 );	//w 31 2 0 1
			RTL8169_WRITE_GMII_REG( ioaddr, 21, 0x1000 );	//w 21 15 0 1000
			RTL8169_WRITE_GMII_REG( ioaddr, 24, 0x65c7 );	//w 24 15 0 65c7
			RTL8169_WRITE_GMII_REG_BIT( ioaddr, 4, 11, 0 );	//w 4 11 11 0
			val = RTL8169_READ_GMII_REG( ioaddr, 4 ) & 0x0fff;
			RTL8169_WRITE_GMII_REG( ioaddr, 4, val );	//w 4 15 12 0
			RTL8169_WRITE_GMII_REG( ioaddr, 3, 0x00a1 );	//w 3 15 0 00a1
			RTL8169_WRITE_GMII_REG( ioaddr, 2, 0x0008 );	//w 2 15 0 0008
			RTL8169_WRITE_GMII_REG( ioaddr, 1, 0x1020 );	//w 1 15 0 1020
			RTL8169_WRITE_GMII_REG( ioaddr, 0, 0x1000 );	//w 0 15 0 1000
			RTL8169_WRITE_GMII_REG_BIT( ioaddr, 4, 11, 1 );	//w 4 11 11 1
			RTL8169_WRITE_GMII_REG_BIT( ioaddr, 4, 11, 0 );	//w 4 11 11 0
			val = ( RTL8169_READ_GMII_REG( ioaddr, 4 ) & 0x0fff ) | 0x7000;
			RTL8169_WRITE_GMII_REG( ioaddr, 4, val );	//w 4 15 12 7
			RTL8169_WRITE_GMII_REG( ioaddr, 3, 0xff41 );	//w 3 15 0 ff41
			RTL8169_WRITE_GMII_REG( ioaddr, 2, 0xde60 );	//w 2 15 0 de60
			RTL8169_WRITE_GMII_REG( ioaddr, 1, 0x0140 );	//w 1 15 0 0140
			RTL8169_WRITE_GMII_REG( ioaddr, 0, 0x0077 );	//w 0 15 0 0077
			RTL8169_WRITE_GMII_REG_BIT( ioaddr, 4, 11, 1 );	//w 4 11 11 1
			RTL8169_WRITE_GMII_REG_BIT( ioaddr, 4, 11, 0 );	//w 4 11 11 0
			val = ( RTL8169_READ_GMII_REG( ioaddr, 4 ) & 0x0fff ) | 0xa000;
			RTL8169_WRITE_GMII_REG( ioaddr, 4, val );	//w 4 15 12 a
			RTL8169_WRITE_GMII_REG( ioaddr, 3, 0xdf01 );	//w 3 15 0 df01
			RTL8169_WRITE_GMII_REG( ioaddr, 2, 0xdf20 );	//w 2 15 0 df20
			RTL8169_WRITE_GMII_REG( ioaddr, 1, 0xff95 );	//w 1 15 0 ff95
			RTL8169_WRITE_GMII_REG( ioaddr, 0, 0xfa00 );	//w 0 15 0 fa00
			RTL8169_WRITE_GMII_REG_BIT( ioaddr, 4, 11, 1 );	//w 4 11 11 1
			RTL8169_WRITE_GMII_REG_BIT( ioaddr, 4, 11, 0 );	//w 4 11 11 0
			val = ( RTL8169_READ_GMII_REG( ioaddr, 4 ) & 0x0fff ) | 0xb000;
			RTL8169_WRITE_GMII_REG( ioaddr, 4, val );	//w 4 15 12 b
			RTL8169_WRITE_GMII_REG( ioaddr, 3, 0xff41 );	//w 3 15 0 ff41
			RTL8169_WRITE_GMII_REG( ioaddr, 2, 0xde20 );	//w 2 15 0 de20
			RTL8169_WRITE_GMII_REG( ioaddr, 1, 0x0140 );	//w 1 15 0 0140
			RTL8169_WRITE_GMII_REG( ioaddr, 0, 0x00bb );	//w 0 15 0 00bb
			RTL8169_WRITE_GMII_REG_BIT( ioaddr, 4, 11, 1 );	//w 4 11 11 1
			RTL8169_WRITE_GMII_REG_BIT( ioaddr, 4, 11, 0 );	//w 4 11 11 0
			val = ( RTL8169_READ_GMII_REG( ioaddr, 4 ) & 0x0fff ) | 0xf000;
			RTL8169_WRITE_GMII_REG( ioaddr, 4, val );	//w 4 15 12 f
			RTL8169_WRITE_GMII_REG( ioaddr, 3, 0xdf01 );	//w 3 15 0 df01
			RTL8169_WRITE_GMII_REG( ioaddr, 2, 0xdf20 );	//w 2 15 0 df20
			RTL8169_WRITE_GMII_REG( ioaddr, 1, 0xff95 );	//w 1 15 0 ff95
			RTL8169_WRITE_GMII_REG( ioaddr, 0, 0xbf00 );	//w 0 15 0 bf00
			RTL8169_WRITE_GMII_REG_BIT( ioaddr, 4, 11, 1 );	//w 4 11 11 1
			RTL8169_WRITE_GMII_REG_BIT( ioaddr, 4, 11, 0 );	//w 4 11 11 0
			RTL8169_WRITE_GMII_REG( ioaddr, 31, 0x0000 );	//w 31 2 0 0
	}
}




#define rtl8169_request_timer( timer, timer_expires, timer_func, timer_data ) \
{ \
	init_timer(timer); \
	timer->expires = (unsigned long)(jiffies + timer_expires); \
	timer->data = (unsigned long)(timer_data); \
	timer->function = (void *)(timer_func); \
	add_timer(timer); \
	DBG_PRINT("request_timer at 0x%08lx\n", (unsigned long)timer); \
}

#define rtl8169_delete_timer( del_timer_t ) \
{ \
	del_timer(del_timer_t); \
	DBG_PRINT("delete_timer at 0x%08lx\n", (unsigned long)del_timer_t); \
}

#define rtl8169_mod_timer( timer, timer_expires ) \
{ \
	mod_timer( timer, jiffies + timer_expires ); \
}




//======================================================================================================
//======================================================================================================
void rtl8169_phy_timer_t_handler( void	*timer_data )
{
	struct net_device *dev = (struct net_device *)timer_data;
	struct rtl8169_private *priv = (struct rtl8169_private *) (dev->priv);
	long ioaddr = (long)(priv->mmio_addr);


	if( RTL_R8(PHYstatus) & LinkStatus ){
		priv->phy_link_down_cnt = 0 ;
	}
	else{
		priv->phy_link_down_cnt ++ ;
		if( priv->phy_link_down_cnt >= 12 ){
			// If link on 1000, perform phy reset.
			if( RTL8169_READ_GMII_REG( ioaddr, PHY_1000_CTRL_REG ) & PHY_Cap_1000_Full )
			{
				rtl8169_hw_PHY_reset( dev );
			}

			priv->phy_link_down_cnt = 0 ;
		}
	}

	//---------------------------------------------------------------------------
	//mod_timer is a more efficient way to update the expire field of an active timer.
	//---------------------------------------------------------------------------
	rtl8169_mod_timer( (&priv->phy_timer_t), 100 );
}




//======================================================================================================
static void *rtl8169_probe1(struct pci_dev *pdev, void *init_dev,
							long ioaddr, int irq, int chip_idx, int found_cnt)
{
	struct net_device *dev;
	struct rtl8169_private *priv;
	void *priv_mem;
	int i;
	int option = -1, Cap10_100 = 0, Cap1000 = 0;
	static int board_idx = -1;

	board_idx++;

	dev = init_etherdev(init_dev, 0);

	for(i=0;i<6;i++){
		dev->dev_addr[i] = RTL_R8(MAC0+i);
	}


	// Make certain elements e.g. descriptor lists are aligned. //
	priv_mem = kmalloc(sizeof(*priv) + PRIV_ALIGN, GFP_KERNEL);
	/* Check for the very unlikely case of no memory. */
	if (priv_mem == NULL)
		return NULL;

	/* We do a request_region() to register /proc/ioports info. */
	request_region(ioaddr, pci_tbl[chip_idx].io_size, dev->name);

	dev->base_addr = ioaddr;
	dev->irq = irq;

	dev->priv = priv = (void *)(((long)priv_mem + PRIV_ALIGN) & ~PRIV_ALIGN);
	memset(priv, 0, sizeof(*priv));
	priv->priv_addr = priv_mem;

	priv->next_module = root_rtl8169_dev;
	root_rtl8169_dev = dev;

	priv->pci_dev = pdev;
	priv->chip_id = chip_idx;

	// The Rtl8169-specific entries in the device structure.
	dev->open = &rtl8169_open;
	dev->hard_start_xmit = &rtl8169_start_xmit;
	dev->stop = &rtl8169_close;
	dev->get_stats = &rtl8169_get_stats;
	dev->set_multicast_list = &set_rx_mode;
	//========================================================================
	// Soft reset the chip.
	rtl8169_soft_reset(dev);

	// identify chip attached to board
	RTL_GET_MAC_VERSION(priv->mac_version);
	RTL_GET_PHY_VERSION(priv->phy_version);

	RTL_PRINT_MAC_VERSION(priv->mac_version);
	RTL_PRINT_PHY_VERSION(priv->phy_version);


	printk (KERN_INFO "%s: PCI space at 0x%lx, MAC=%02x:%02x:%02x:%02x:%02x:%02x, IRQ=%d\n",
				dev->name,
				dev->base_addr,
				dev->dev_addr[0], dev->dev_addr[1],
				dev->dev_addr[2], dev->dev_addr[3],
				dev->dev_addr[4], dev->dev_addr[5],
				dev->irq);


	// Config PHY
	rtl8169_hw_PHY_config(dev);

	DBG_PRINT("Set MAC Reg C+CR Offset 0x82h = 0x01h\n");
	RTL_W8( 0x82, 0x01 );


	if( priv->mac_version < RTL_GIGA_MAC_VER_E ){
		DBG_PRINT("Set PCI Latency=0x40\n");
		pci_write_config_byte(pdev, PCI_LATENCY_TIMER, 0x40);
	}

	if( priv->mac_version == RTL_GIGA_MAC_VER_D ){
		DBG_PRINT("Set MAC Reg C+CR Offset 0x82h = 0x01h\n");
		RTL_W8( 0x82, 0x01 );
		DBG_PRINT("Set PHY Reg 0x0bh = 0x00h\n");
		RTL8169_WRITE_GMII_REG( ioaddr, 0x0b, 0x0000 );	//w 0x0b 15 0 0
	}

	//========================================================================
	//Set Link Type !
	//if(!TBI_Link)
	//	if(Force_Mode_Enabled){set auto-negotiation ability to the forced speed}
	//	else {set all of the auto-negotiation ability}
	//else
	//	Just check TBI link status
	//========================================================================
	if( !(RTL_R8(PHYstatus) & TBI_Enable) ){
		int	val = RTL8169_READ_GMII_REG( ioaddr, PHY_AUTO_NEGO_REG );

		option = media[board_idx];
		// Force RTL8169 in 10/100/1000 Full/Half mode.
		if( option > 0 ){
			printk(KERN_INFO "%s: Force-mode Enabled.\n", dev->name);
			switch( option ){
				case _10_Half: 	Cap10_100 = PHY_Cap_10_Half;	Cap1000 = PHY_Cap_Null;		break;
				case _10_Full: 	Cap10_100 = PHY_Cap_10_Full;	Cap1000 = PHY_Cap_Null;		break;
				case _100_Half: Cap10_100 = PHY_Cap_100_Half;	Cap1000 = PHY_Cap_Null;		break;
				case _100_Full: Cap10_100 = PHY_Cap_100_Full;	Cap1000 = PHY_Cap_Null; 	break;
				case _1000_Full:Cap10_100 = PHY_Cap_Null;	Cap1000 = PHY_Cap_1000_Full;	break;
				default: break;
			}
			RTL8169_WRITE_GMII_REG( ioaddr, PHY_AUTO_NEGO_REG, Cap10_100 | ( val&0x1F ) );
			RTL8169_WRITE_GMII_REG( ioaddr, PHY_1000_CTRL_REG, Cap1000 );
		}
		else{
			printk(KERN_INFO "%s: Auto-negotiation Enabled.\n", dev->name);
	
			// enable 10/100 Full/Half Mode
			RTL8169_WRITE_GMII_REG( ioaddr, PHY_AUTO_NEGO_REG, PHY_Cap_10_Half | PHY_Cap_10_Full | PHY_Cap_100_Half | PHY_Cap_100_Full | ( val&0x1F ) );
	
			// enable 1000 Full Mode
			RTL8169_WRITE_GMII_REG( ioaddr, PHY_1000_CTRL_REG, PHY_Cap_1000_Full );
	
		}// end of if( option > 0 )

		// Enable auto-negotiation and restart auto-nigotiation
		RTL8169_WRITE_GMII_REG( (long)ioaddr, PHY_CTRL_REG, PHY_Enable_Auto_Nego | PHY_Restart_Auto_Nego );

		// wait for auto-negotiation process
		for( i = 2000; i > 0; i-- ){
			//check if auto-negotiation complete
			if( RTL8169_READ_GMII_REG(ioaddr, PHY_STAT_REG) & PHY_Auto_Neco_Comp ){
				udelay(100);
				option = RTL_R8(PHYstatus);
				if( option & _1000bpsF ){
					printk(KERN_INFO "%s: 1000Mbps Full-duplex Link %s!\n",
						dev->name,
						( RTL_R8(PHYstatus)&LinkStatus )?"OK":"Failed");
				}
				else{
					printk(KERN_INFO "%s: %sMbps %s-duplex Link %s!\n",
						dev->name,
						(option & _100bps) ? "100" : "10",
						(option & FullDup) ? "Full" : "Half",
						( RTL_R8(PHYstatus)&LinkStatus )?"OK":"Failed" );
				}
				break;
			}
			else{
				udelay(100);
			}// end of if( RTL8169_READ_GMII_REG(ioaddr, 1) & 0x20 )
		}// end for-loop to wait for auto-negotiation process

		if(i==0)printk(KERN_INFO"%s: Auto-negotiation process is not completed!\n", dev->name);
	}// end of TBI is not enabled
	else{	
		udelay(100);
		printk(KERN_INFO "%s: 1000Mbps Full-duplex operation, TBI Link %s!\n",
					dev->name, (RTL_R32(TBICSR) & TBILinkOK) ? "OK" : "Failed" );
	}// end of TBI is not enabled
	//========================================================================

	return dev;
}














//======================================================================================================
static int rtl8169_init_ring (struct net_device *dev)
{
	struct rtl8169_private *priv = (struct rtl8169_private *)dev->priv;
	int	diff;
	int i;
	u32 TxPhyAddr, RxPhyAddr;


	priv->cur_rx = 0;
	priv->cur_tx = 0;
	priv->dirty_tx = 0;
	priv->tx_full = 0;

	//===============================================================================
	// Allocate memory for Tx Desscriptors.
	// Make sure memory address to be 256 bytes alignment.
	//===============================================================================
	priv->TxDescArrays = kmalloc( NUM_TX_DESC * sizeof(struct TxDesc)+256 , GFP_KERNEL );
	if ( priv->TxDescArrays == NULL ) {
		printk(KERN_INFO"Allocate TxDescArray failed\n");
		free_irq(dev->irq, dev);
		return -ENOMEM;
	}
	TxPhyAddr = virt_to_bus(priv->TxDescArrays);
	diff = 256 - (TxPhyAddr-((TxPhyAddr >> 8)<< 8));
	priv->TxDescArray = (struct TxDesc *)(priv->TxDescArrays + diff);

	//===============================================================================
	// Allocate memory for Rx Desscriptors.
	// Make sure memory address to be 256 bytes alignment.
	//===============================================================================
	priv->RxDescArrays = kmalloc( NUM_RX_DESC * sizeof(struct RxDesc)+256 , GFP_KERNEL );
	if ( priv->TxDescArrays == NULL || priv->RxDescArrays == NULL ) {
		printk(KERN_INFO"Allocate RxDescArray failed\n");
		free_irq(dev->irq, dev);
		if( priv->TxDescArrays ) kfree(priv->TxDescArrays);
		return -ENOMEM;
	}
	RxPhyAddr = virt_to_bus(priv->RxDescArrays);
	diff = 256 - (RxPhyAddr-((RxPhyAddr >> 8)<< 8));
	priv->RxDescArray = (struct RxDesc *)(priv->RxDescArrays + diff);


	{        int i;
		struct sk_buff *skb = NULL;

		for(i=0;i<NUM_RX_DESC;i++){
			skb = RTL8169_ALLOC_RXSKB(MAX_RX_SKBDATA_SIZE);
			if( skb != NULL ) {
				skb_reserve (skb, 2);	// 16 byte align the IP fields. //
				priv->Rx_skbuff[i] = skb;
			}
			else{
				printk(KERN_INFO"%s: Gigabit driver failed to allocate skbuff.\n", dev->name);
				priv->drvinit_fail = 1;
			}
		}
	}


	//===============================================================================
	// Reset Tx and Rx Desscriptors.
	//===============================================================================
	memset( priv->TxDescArray, 0x0, NUM_TX_DESC *sizeof(struct TxDesc) );
	memset( priv->RxDescArray, 0x0, NUM_RX_DESC *sizeof(struct RxDesc) );

	//===============================================================================
	// Reset Tx skb buffer.
	//===============================================================================
	for (i=0 ; i<NUM_TX_DESC ; i++){
		priv->Tx_skbuff[i] = NULL;
	}

	//===============================================================================
	// Fill Rx descripter in advance
	//===============================================================================
	for (i=0; i <NUM_RX_DESC; i++) {
		if( i == (NUM_RX_DESC-1) )
			priv->RxDescArray[i].status = (OWNbit | EORbit) + RX_BUF_SIZE;
		else
			priv->RxDescArray[i].status = OWNbit + RX_BUF_SIZE;

		{//-----------------------------------------------------------------------
			struct sk_buff *skb = priv->Rx_skbuff[i];

			if( skb != NULL ){
				priv->RxDescArray[i].buf_addr = virt_to_bus( skb->data );
			}
			else{
				DBG_PRINT("%s: %s() Rx_skbuff == NULL\n", dev->name, __FUNCTION__);
				priv->drvinit_fail = 1;
			}
		}//-----------------------------------------------------------------------

	}

	return 0;

}









//======================================================================================================
/* Start the hardware at open or resume. */
static void rtl8169_hw_start (struct net_device *dev)
{

	struct rtl8169_private *priv = (struct rtl8169_private *)dev->priv;
	long ioaddr = dev->base_addr;


	// Soft reset the chip.
	rtl8169_soft_reset(dev);


	priv->cur_rx = 0;

	RTL_W8( Cfg9346, Cfg9346_Unlock );
	RTL_W8( ChipCmd, (CmdRxEnb | CmdTxEnb) );
	RTL_W8( EarlyTxThres, EarlyTxThld );


	// For gigabit rtl8169
    RTL_W16( RxMaxSize, RxPacketMaxSize );

	// Set Rx Config register
	RTL_W32( RxConfig, (rtl8169_rx_config | ( RTL_R32( RxConfig ) & 0xff7e1880 )) );

	priv->rx_config = rtl8169_rx_config;

	// Set DMA burst size and Interframe Gap Time //
	RTL_W32( TxConfig, rtl8169_tx_config );


	//2003-07-18
	if( priv->mac_version == RTL_GIGA_MAC_VER_D ){
		DBG_PRINT("Set MAC Reg C+CR Offset 0xE0: bit-3 and bit-14 MUST be 1\n");
		RTL_W16( CPlusCmd, (RTL_R16(CPlusCmd)|(1<<14)|(1<<3)) );
	}

	
	RTL_W32( TxDescStartAddr, virt_to_bus(priv->TxDescArray) );
	RTL_W32( RxDescStartAddr, virt_to_bus(priv->RxDescArray) );

	RTL_W8( Cfg9346, Cfg9346_Lock );


	RTL_W32( RxMissed, 0x00 );

	set_rx_mode (dev);


	RTL_W16( MultiIntr, (RTL_R16(MultiIntr) & 0xF000) );

	// Enable all known interrupts by setting the interrupt mask. //
	RTL_W16( IntrMask, rtl8169_intr_mask );

//	netif_start_queue (dev);

}//end of rtl8169_hw_start (struct net_device *dev)









//======================================================================================================
static int rtl8169_open(struct net_device *dev)
{
	struct rtl8169_private *priv = (struct rtl8169_private *)dev->priv;


	MOD_INC_USE_COUNT;
	if (request_irq(dev->irq, &rtl8169_interrupt, SA_SHIRQ, dev->name, dev)) {
		MOD_DEC_USE_COUNT;
		return -EAGAIN;
	}

	rtl8169_init_ring (dev);
	rtl8169_hw_start (dev);

	// ------------------------------------------------------
	if( (priv->mac_version > RTL_GIGA_MAC_VER_B) && (priv->phy_version < RTL_GIGA_PHY_VER_G) ){
		DBG_PRINT("FIX PCS -> rtl8169_request_timer\n");
		rtl8169_request_timer( (&priv->phy_timer_t), 100, rtl8169_phy_timer_t_handler, ((void *)dev) );  //in open()
		priv->phy_link_down_cnt = 0;
	}

	dev->tbusy = 0;


	return 0;
}








//======================================================================================================
static int rtl8169_start_xmit(struct sk_buff *skb, struct net_device *dev)
{
	struct rtl8169_private *priv = (struct rtl8169_private *)dev->priv;
	long ioaddr = dev->base_addr;
	int entry = priv->cur_tx % NUM_TX_DESC;


	if (test_and_set_bit(0, (void*)&dev->tbusy) != 0) {
		if (jiffies - dev->trans_start >= TX_TIMEOUT){
			//-------------------------------//
			// reset chip and restart driver //
			//-------------------------------//
			//rtl8129_tx_timeout(dev);
		}
		return 1;
	}


	if( (priv->TxDescArray[entry].status & OWNbit)==0 ){
		priv->Tx_skbuff[entry] = skb;
		priv->TxDescArray[entry].buf_addr = virt_to_bus(skb->data);
		if( entry != (NUM_TX_DESC-1) ){
			priv->TxDescArray[entry].status = (OWNbit | FSbit | LSbit) | ( (skb->len > ETH_ZLEN) ? skb->len : ETH_ZLEN);
		}
        	else{
		 	priv->TxDescArray[entry].status = (OWNbit | EORbit | FSbit | LSbit) | ( (skb->len > ETH_ZLEN) ? skb->len : ETH_ZLEN);
		}

		RTL_W8( TxPoll, 0x40 );	//set polling bit

		if (++priv->cur_tx - priv->dirty_tx < NUM_TX_DESC){
			clear_bit(0, (void*)&dev->tbusy);  /* Typical path */
		}
		else {
			set_bit(0, &priv->tx_full);
			if (priv->cur_tx - (volatile int)priv->dirty_tx < NUM_TX_DESC) {
				clear_bit(0, (void*)&dev->tbusy);
				clear_bit(0, &priv->tx_full);
			}
		}

		dev->trans_start = jiffies;

	}//end of if( (priv->TxDescArray[entry].status & 0x80000000)==0 )

	return 0;
}








//======================================================================================================
static void rtl8169_tx_interrupt (struct net_device *dev, struct rtl8169_private *priv, long ioaddr)
{

	unsigned long tx_left=0, dirty_tx = priv->dirty_tx;
	int entry = priv->cur_tx % NUM_TX_DESC;


	tx_left = priv->cur_tx - dirty_tx;

	if( tx_left > 20 ){
		tx_left = 20;
	}

	while (tx_left > 0) {					
		if( (priv->TxDescArray[entry].status & OWNbit) == 0 ){
			dev_free_skb(priv->Tx_skbuff[dirty_tx % NUM_TX_DESC]);
			priv->Tx_skbuff[dirty_tx % NUM_TX_DESC] = NULL;
			if (test_bit(0, &priv->tx_full)) {
				/* The ring is no longer full, clear tbusy. */
				clear_bit(0, &priv->tx_full);
				clear_bit(0, (void*)&dev->tbusy);
				netif_wake_queue(dev);
			}
			priv->stats.tx_packets++;			
			dirty_tx++;				
			tx_left--;
			entry++;
		}
	}						

	if (priv->dirty_tx != dirty_tx) {			
		priv->dirty_tx = dirty_tx;			
		netif_wake_queue (dev);
	}							

}







//======================================================================================================
static void rtl8169_rx_interrupt (struct net_device *dev, struct rtl8169_private *priv, long ioaddr)
{

	int pkt_size = 0 ;
	int rxdesc_cnt = 0;
	int cur_rx = priv->cur_rx;
	struct sk_buff *n_skb = NULL;
	struct sk_buff *rx_skb = priv->Rx_skbuff[cur_rx];
	struct sk_buff *cur_skb;



	while ( ((priv->RxDescArray[cur_rx].status & OWNbit)==0) && (rxdesc_cnt < 20) ) {

		rxdesc_cnt++;

		if( priv->RxDescArray[cur_rx].status & RxRES ){
			printk(KERN_INFO "%s: Rx ERROR!!!\n", dev->name);
			priv->stats.rx_errors++;
			if (priv->RxDescArray[cur_rx].status & (RxRWT|RxRUNT) ){
				priv->stats.rx_length_errors++;
			}
			if (priv->RxDescArray[cur_rx].status & RxCRC){
				priv->stats.rx_crc_errors++;
			}
	    }
	    else{
			pkt_size=(int)(priv->RxDescArray[cur_rx].status & 0x00001FFF)-4;

			{// -----------------------------------------------------
				rx_skb = priv->Rx_skbuff[cur_rx];

				//n_skb = NULL;
				n_skb = RTL8169_ALLOC_RXSKB(MAX_RX_SKBDATA_SIZE);
				if( n_skb != NULL ) {
					skb_reserve (n_skb, 2);	// 16 byte align the IP fields. //

					// Indicate rx_skb
					if( rx_skb != NULL ){
						rx_skb->dev = dev;
						skb_put ( rx_skb, pkt_size );
						rx_skb->protocol = eth_type_trans ( rx_skb, dev );
						RTL8169_NETIF_RX (rx_skb);

						dev->last_rx = jiffies;
						priv->stats.rx_bytes += pkt_size;
						priv->stats.rx_packets++;
					}//end if( rx_skb != NULL )

					priv->Rx_skbuff[cur_rx] = n_skb;
				}
				else{
					priv->Rx_skbuff[cur_rx] = rx_skb;
				}

				// Update rx descriptor
				if( cur_rx == (NUM_RX_DESC-1) ){
					priv->RxDescArray[cur_rx].status  = (OWNbit | EORbit) + RX_BUF_SIZE;
				}
				else{
					priv->RxDescArray[cur_rx].status  = OWNbit + RX_BUF_SIZE;
				}

				cur_skb = priv->Rx_skbuff[cur_rx];
				if( cur_skb != NULL ){
					priv->RxDescArray[cur_rx].buf_addr = virt_to_bus( cur_skb->data );
				}
				else{
					DBG_PRINT("%s: %s() cur_skb == NULL\n", dev->name, __FUNCTION__);
				}

			}//------------------------------------------------------------

		}//end if( priv->RxDescArray[cur_rx].status & RxRES ){

	    cur_rx = (cur_rx +1) % NUM_RX_DESC;

	}// end of while ( (priv->RxDescArray[cur_rx].status & 0x80000000)== 0)

	priv->cur_rx = cur_rx;
}






//======================================================================================================
/* The interrupt handler does all of the Rx thread work and cleans up after the Tx thread. */
static void rtl8169_interrupt(int irq, void *dev_instance, struct pt_regs *regs)
{
	struct net_device *dev = (struct net_device *) dev_instance;
	struct rtl8169_private *priv = (struct rtl8169_private *)dev->priv;
	int boguscnt = max_interrupt_work;
	long ioaddr = dev->base_addr;
	int status = 0;


	do {
		status = RTL_R16(IntrStatus);

		// h/w no longer present (hotplug?) or major error, bail //
		if (status == 0xFFFF)
			break;

		RTL_W16( IntrStatus, status );

		if ( (status & rtl8169_intr_mask) == 0 )
			break;


		// Rx interrupt
//		if (status & (RxOK | RxUnderrun | RxOverflow | RxFIFOOver)){
			rtl8169_rx_interrupt (dev, priv, ioaddr);
//		}

		// Tx interrupt
//		if (status & (TxOK | TxErr)) {
			rtl8169_tx_interrupt (dev, priv, ioaddr);
//		}

		boguscnt--;
	} while (boguscnt > 0);

	if (boguscnt <= 0) {
		printk (KERN_WARNING "%s: Too much work at interrupt!\n", dev->name);
		// Clear all interrupt sources.//
		RTL_W16( IntrStatus, 0xffff );
	}

	return;
}







//======================================================================================================
static void rtl8169_tx_clear (struct rtl8169_private *tp)
{
	int i;


	tp->cur_tx = 0;
	for ( i = 0 ; i < NUM_TX_DESC ; i++ ){
		if ( tp->Tx_skbuff[i] != NULL ) {
			dev_kfree_skb ( tp->Tx_skbuff[i] );
			tp->Tx_skbuff[i] = NULL;
			tp->stats.tx_dropped++;
		}
	}

}






//======================================================================================================
static int rtl8169_close(struct net_device *dev)
{

	struct rtl8169_private *priv = (struct rtl8169_private *)dev->priv;
	long ioaddr = dev->base_addr;
	int i;


	// ------------------------------------------------------
	if( (priv->mac_version > RTL_GIGA_MAC_VER_B) && (priv->phy_version < RTL_GIGA_PHY_VER_G) ){
		DBG_PRINT("FIX PCS -> rtl8169_delete_timer\n");
		rtl8169_delete_timer( &(priv->phy_timer_t) ); //in close()
		priv->phy_link_down_cnt = 0;
	}


	dev->tbusy = 1;

	// Stop the chip's Tx and Rx DMA processes. //
    RTL_W8( ChipCmd, 0x00 );

	// Disable interrupts by clearing the interrupt mask. //
	RTL_W16( IntrMask, 0x0000 );

	// Update the error counts. //
	priv->stats.rx_missed_errors += RTL_R32(RxMissed);
	RTL_W32( RxMissed, 0x00000000 );

	free_irq (dev->irq, dev);

	rtl8169_tx_clear (priv);
	kfree(priv->TxDescArrays);
	kfree(priv->RxDescArrays);
	priv->TxDescArrays = NULL;
	priv->RxDescArrays = NULL;
	priv->TxDescArray = NULL;
	priv->RxDescArray = NULL;

	{//-----------------------------------------------------------------------------
		for(i=0;i<NUM_RX_DESC;i++){
			if( priv->Rx_skbuff[i] != NULL ) {
				RTL8169_FREE_RXSKB ( priv->Rx_skbuff[i] );
			}
		}
	}//-----------------------------------------------------------------------------

	MOD_DEC_USE_COUNT;
	return 0;
}






//======================================================================================================
static struct net_device_stats *rtl8169_get_stats(struct net_device *dev)
{
	struct rtl8169_private *tp = (struct rtl8169_private *)dev->priv;
	long ioaddr = dev->base_addr;

	if (dev->start) {
		tp->stats.rx_missed_errors += RTL_R32(RxMissed);
		RTL_W32( RxMissed, 0x00000000 );
	}

	return &tp->stats;
}





//======================================================================================================
/* Set or clear the multicast filter for this adaptor.
   This routine is not state sensitive and need not be SMP locked. */

static unsigned const ethernet_polynomial = 0x04c11db7U;
static inline u32 ether_crc(int length, unsigned char *data)
{
    int crc = -1;

    while (--length >= 0) {
		unsigned char current_octet = *data++;
		int bit;
		for (bit = 0; bit < 8; bit++, current_octet >>= 1)
			crc = (crc << 1) ^
				((crc < 0) ^ (current_octet & 1) ? ethernet_polynomial : 0);
    }
    return crc;
}







//======================================================================================================
static void set_rx_mode(struct net_device *dev)
{
	struct rtl8169_private *tp = (struct rtl8169_private *)dev->priv;
	long ioaddr = dev->base_addr;
	u32 mc_filter[2];		 /* Multicast hash filter */
	int i, rx_mode;



	/* Note: do not reorder, GCC is clever about common statements. */
	if (dev->flags & IFF_PROMISC) {
		/* Unconditionally log net taps. */
		printk(KERN_NOTICE"%s: Promiscuous mode enabled.\n", dev->name);
		rx_mode = AcceptBroadcast|AcceptMulticast|AcceptMyPhys|AcceptAllPhys;
		mc_filter[1] = mc_filter[0] = 0xffffffff;
	} else if ((dev->mc_count > multicast_filter_limit)
			   || (dev->flags & IFF_ALLMULTI)) {
		/* Too many to filter perfectly -- accept all multicasts. */
		rx_mode = AcceptBroadcast | AcceptMulticast | AcceptMyPhys;
		mc_filter[1] = mc_filter[0] = 0xffffffff;
	} else {
		struct dev_mc_list *mclist;
		rx_mode = AcceptBroadcast | AcceptMulticast | AcceptMyPhys;
		mc_filter[1] = mc_filter[0] = 0;
		for (i = 0, mclist = dev->mc_list; mclist && i < dev->mc_count;
			 i++, mclist = mclist->next)
			set_bit(ether_crc(ETH_ALEN, mclist->dmi_addr) >> 26, mc_filter);
	}
	
	/* We can safely update without stopping the chip. */
	RTL_W32( RxConfig, (tp->rx_config | rx_mode) );
	RTL_W32( (MAR0 + 0), mc_filter[0] );
	RTL_W32( (MAR0 + 4), mc_filter[1] );


	return;
}



#ifdef MODULE
//======================================================================================================
int init_module(void)
{
	printk (KERN_INFO RTL8169_DRIVER_NAME " loaded\n");
	return pci_drv_register(&rtl8169_drv_id, NULL);
}


//======================================================================================================
void cleanup_module(void)
{
	struct net_device *next_dev;

	pci_drv_unregister(&rtl8169_drv_id);

	while (root_rtl8169_dev) {
		struct rtl8169_private *np = (struct rtl8169_private *)(root_rtl8169_dev->priv);
		unregister_netdev(root_rtl8169_dev);
		release_region(root_rtl8169_dev->base_addr, pci_tbl[np->chip_id].io_size);
#ifndef USE_IO_OPS
		iounmap((char *)(root_rtl8169_dev->base_addr));
#endif
		next_dev = np->next_module;
		if (np->priv_addr)
			kfree(np->priv_addr);
		kfree(root_rtl8169_dev);
		root_rtl8169_dev = next_dev;
	}
}

#endif  /* MODULE */


/*
 * Local variables:
 *  compile-command: "gcc -DMODULE -Wall -Wstrict-prototypes -O6 -c rtl8169.c"
 */
